/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.DT;

import hypercast.*;
import hypercast.adapters.*;

import java.net.*;
import java.io.*;
import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import org.apache.xpath.*;

/** This class implements the DT Protocol server. 
  *<p>
 * @author HyperCast Team
 * @author Michael Nahas
 * @author Weisheng Si
 * @version	2.0, Oct. 16, 2001
 * @author Jianping Wang
 * @version 3.0, Jan. 08, 2004 
 */


final public class DT_Server implements I_AdapterCallback, I_Node
{
	/** Default index of DT overlays supported by DT_server. */
	private static final int DEFAULT_OVERLAYINDEX = 0;

	/** Default cofiguration file name used by DT_server. */
	private static final String DEFAULT_ADDRSTRING = "8081";

	public final static long T_REFRESH = 2000; // milliseconds

	private final static int HEARTBEAT_TIMER_INDEX = 1;

	/** The cache of nodes in the Overlay. */
	private NodeCache nodeCache;

	/** The adapter of the DT server. */
	public I_UnicastAdapter adapter;

	/** Used for sending messages. */
	public DT_AddressPair myAddressPair;

	/** Vector of Long.  Times of requests - ordered from oldest to newest. */
	private Vector requestTimeQueue;

	/** config object used in creation of Adapter */
	private HyperCastConfig config;
    
	/** Statistic. */
	private int numberOfRequests;

	/** Statistic */
	private long timeOfFirstRequest;

	public static Random rand = new Random();

	/** The protocol subfield value which represents the used rendezvous and will be put 
	 * in the Proto_sub field of a DT_Message. 
	 */
	byte Proto_Sub = I_Node.PROTOSUB_SERVER;

	/* Constructor.
	 * @param	addrString	the string used by the adapter to create
	 * 					physical address. It may only contains a port
	 * 					string representing the port that this DT server
	 * 					will listen to.
	 * @param	config_f	configuration file name
	 */
	
	DT_Server(String addrString, String config_f)
	{
		Initialization(config_f);

		AdapterFactory adapterF= new AdapterFactory();

		adapter = adapterF.createAdapter(config, null, "NodeAdapter", addrString);

		adapter.setCallback(this);
		adapter.Start();

		adapter.startTimer();
		adapter.setTimer(new Integer(HEARTBEAT_TIMER_INDEX), (long) (rand.nextDouble() * T_REFRESH)); 

		I_PhysicalAddress myIPAddr = (I_PhysicalAddress) adapter.createPhysicalAddress();
		myAddressPair = new DT_AddressPair(myIPAddr, null);

		// print message to user
		System.out.println("DT Server Starts Successfully at " + myIPAddr.toString() + " ...");
	}

	/* Constructor.
	 * @param	Adapter	an I_UnicastAdapter object.
	 * @param	config_file	configuration file name
	 */
	public DT_Server(HyperCastConfig config, I_UnicastAdapter Adapter){
		nodeCache = new NodeCache();

		requestTimeQueue = new Vector();
		numberOfRequests = 0;
		timeOfFirstRequest = -1;

		this.config=config;

		this.adapter = Adapter;
		this.adapter.setCallback(this);
		this.config = config;
		
		adapter.Start();

		adapter.startTimer();
		adapter.setTimer(new Integer(HEARTBEAT_TIMER_INDEX), (long) (rand.nextDouble() * T_REFRESH));
		
		I_PhysicalAddress myIPAddr = (I_PhysicalAddress) adapter.createPhysicalAddress();
		myAddressPair = new DT_AddressPair(myIPAddr, null);
		
		//		 print message to user
		System.out.println("DT Server Starts Successfully at " + myIPAddr.toString() + " ...");
		
	}
	DT_Server(I_UnicastAdapter Adapter, String config_f) 
	{
		Initialization(config_f);

		adapter = Adapter;

		adapter.setCallback(this);
		adapter.Start();

		adapter.startTimer();
		adapter.setTimer(new Integer(HEARTBEAT_TIMER_INDEX), (long) (rand.nextDouble() * T_REFRESH)); 

		I_PhysicalAddress myIPAddr = (I_PhysicalAddress) adapter.createPhysicalAddress();
		myAddressPair = new DT_AddressPair(myIPAddr, null);

		// print message to user
		System.out.println("DT Server Starts Successfully at " + myIPAddr.toString() + " ...");
	}


	/* Finish the common initialization work for constructors.
	 * @param	config_file	configuration file name
	 */
	void Initialization(String config_file) 
	{
		nodeCache = new NodeCache();

		requestTimeQueue = new Vector();
		numberOfRequests = 0;
		timeOfFirstRequest = -1;

		config=null;

		try 
		{
			config = HyperCastConfig.createConfig(config_file);
		}
		catch (Exception e) 
		{
			throw new HyperCastFatalRuntimeException ("Error occured while trying to create overlay.", e);
		}

	}


	/** return the Overlay Hash */
	public int getOverlayHash() 
	{
		return config.getOverlayHash();
	}


	/** handle messages when they arrived at the Message Buffer. */
	public synchronized void messageArrivedFromAdapter(I_Message m) 
	{
		DT_Message response = handleMessage((DT_Message) m);
		if (response != null)
			adapter.sendUnicastMessage(response.getDstPA(), response);
	}
    
	/** Process received message based on the message types. */
	public synchronized DT_Message handleMessage(DT_Message m) 
	{
		if (m == null)
			return null;

		long currentTime = System.currentTimeMillis();
		if (timeOfFirstRequest < 0)
			timeOfFirstRequest = currentTime;
		requestTimeQueue.addElement(new Long(currentTime));
		numberOfRequests++;

		switch(m.getType()) 
		{   
		  case DT_Message.NodePong:
			return handleNodePong(m);

		  case DT_Message.ServerRequest:
			return handleServerRequest(m);
            
		  case DT_Message.Goodbye: 
			return handleGoodbye(m);

		  default:
			config.err.println("DT_Node: message with unknown type arrived!");
			return null;
		}
	}


	/** handle serverRequest messages */
	private DT_Message handleServerRequest(DT_Message m) 
	{
		nodeCache.update(m.getSrc());

		DT_AddressPair aParentNode = nodeCache.getAParent(m.getSrc());
        
		DT_Message NewPacket = createDTMessage(DT_Message.ServerReply, m.getSrc(), aParentNode, null);

		//System.out.println("OL_Server:handleServerRequest: received msg from:" + ((DT_AddressPair) m.getSrc()).toString() + " index=" + m.getIndex());

		return NewPacket;
	}


	/** handle Goodbye messages */
	private DT_Message handleGoodbye(DT_Message m) 
	{
		nodeCache.remove(m.getSrcPA());
		return null;
	}

	/** handle CachePong messages */
	private DT_Message handleNodePong(DT_Message m) 
	{
		nodeCache.update(m.getSrc());
		return null;
	}

	/** restore a message from a byte array */
	public I_Message restoreMessage(byte[] buffer, int[] startValidBytes, int endValidBytes) 
	{
		return DT_Message.restoreMessage(buffer, startValidBytes, endValidBytes, adapter, config.getOverlayHash());
	}
    
	/** This method is called when timer expires */
	public synchronized void timerExpired(Object Timer_Index) 
	{

		int timerIndex = ((Integer) Timer_Index).intValue();

		if (timerIndex == HEARTBEAT_TIMER_INDEX) 
		{ 
			adapter.setTimer(new Integer(HEARTBEAT_TIMER_INDEX), T_REFRESH);

			// remove timed-out nodes
			nodeCache.removeOldRecords();

            
			for (int i = 0; i < nodeCache.size(); i++)  
			{
				NodeRecord record = nodeCache.elementAt(i);

				// send NodePing message to alive nodes
				DT_Message NewPacket = createDTMessage(DT_Message.NodePing, record.dtap, null, null);

				adapter.sendUnicastMessage(record.dtap.getPhysicalAddress(), NewPacket);
			}
		}
	}

  /* This method creates a new DT_Message based on parameters. 
   * It is called by all sending methods and can be overrided in secure node class.
   */
  public DT_Message createDTMessage (byte type, DT_AddressPair dst, DT_AddressPair cw, DT_AddressPair ccw) {
		
	  return new DT_Message(Proto_Sub,
							type,
							config.getOverlayHash(),
							myAddressPair,
							dst,
							cw,
							ccw);
  }


    /** Start the DT server.
    */
    public static void main(String[] args)
    {

        if (args.length > 3)
        {
            throw new HyperCastFatalRuntimeException ("Too many arguments!" + 
                                                      "USAGE: java -classpath hypercast.jar hypercast.DT.DT_Server" + 
                                                      "            [addrportString|port#] [config_file]");
        }        

        if (args.length == 2) {
			DT_Server dtServer = new DT_Server(args[0], args[1]);
        } else if (args.length == 1) {
			DT_Server dtServer = new DT_Server(args[0], DEFAULT_CONFIG_FILE);
        } else {
			DT_Server dtServer = new DT_Server(DEFAULT_ADDRSTRING, DEFAULT_CONFIG_FILE);
        }        
    }


	/** Joins the overlay.  Do nothing */
	public void joinOverlay(){
	}

	/** Leaves the overlay.  Do nothing. */
	public void leaveOverlay(){
	}
	
	
	/** Returns the addresspair of the next hop for a message routed by this node
	 * towards the root.  Return null;
	 */
	public I_AddressPair[] getParent(I_LogicalAddress root) {return null;}

	/** Returns the node's children's physical/logical address pairs, with respect to
	 * the spanning tree rooted at <code>root</code>. Return null;
	 */
	public I_AddressPair[] getChildren(I_LogicalAddress root) {return null;}

	/** Returns the node's neighbors' physical/logical address pairs.
	 * Return null;
	 */
	public I_AddressPair[] getAllNeighbors() {return null;}
    
	/** Returns this logical and physical addresses of this node.
	 */
	public I_AddressPair getAddressPair() {return myAddressPair;}
	
  /** Returns this physical addresses of this node. 
	   */
	public I_PhysicalAddress getPhysicalAddress()
	{
	   return getAddressPair().getPhysicalAddress();
	}

	/** Returns this logical addresses of this node. 
		 */
	public I_LogicalAddress getLogicalAddress()
	{
		return getAddressPair().getLogicalAddress();
	}

	/** Creates a logical address object from a byte array. Return null.
	 */
	public I_LogicalAddress createLogicalAddress(byte[] laddr, int offset) {return null;}

  /** Creates a logical address object from a String. Return null.
   */
  public I_LogicalAddress createLogicalAddress(String laStr)
  {
	  return null;

  }


  /** Sets the logical address to specified one. Do nothing.
   */
  public void setLogicalAddress(I_LogicalAddress la)
  {
  }


	/** Returns the addresspair of the next hop corresponding to the node with specified index 
	 * for a message routed by this node towards the root.  Return null.
	 */
	public I_AddressPair[] getParent(I_LogicalAddress root, int index) {return null;}

	/** Returns the node's, which has specified index, children's physical/logical 
	 * address pairs , with respect to the spanning tree rooted at <code>root</code>.
	 * Return null..
	 */
	public I_AddressPair[] getChildren(I_LogicalAddress root, int index) {return null;}

	/** Returns the node's ,which has specified index, neighbors' physical/logical 
	 * address pairs. Return null.
	 */
	public I_AddressPair[] getAllNeighbors(int index) {return null;}
    
	/** Returns the node's neighbors' physical/logical address pairs. 
	 * The logical address root contains index information. Return null.
	 */
	public I_AddressPair[] getAllNeighbors(I_LogicalAddress root) {return null;}

	/** Returns this logical and physical addresses of the node with specific index. 
	 */
	public I_AddressPair getMyAddressPair(int index) {return myAddressPair;}
	
  /** Returns the physical address/logical address pair of this node. 
   * The logical address is ignored.
   */
  public synchronized I_AddressPair getMyAddressPair(I_LogicalAddress root)
  {
	  return myAddressPair;
  }

  /** Returns this physical addresses of this node. Index is ignored.
	   */
	public I_PhysicalAddress getPhysicalAddress(int index)
	{
	   return getPhysicalAddress();
	}

	/** Returns this logical addresses of this node. Index is ignored.
		 */
	public I_LogicalAddress getLogicalAddress(int index)
	{
		return getLogicalAddress();
	}

	/** Creates a logical address object based on the specified index from a byte array.
	 * Return null.
	 */
	public I_LogicalAddress createLogicalAddress(byte[] laddr, int offset, int index) {return null;}


	/** Verify the previous hop of the message. If checkmode is set to "neighborCheck",
	 * this method checks if the previous hop is a neighbor; otherwise this method checks
	 * if the previous hop is a valid sender of OL messages. It does nothing in DT_server.
	 */
	public boolean previousHopCheck(I_LogicalAddress src, I_LogicalAddress dst, I_LogicalAddress prehop)
	{
	   return true;
	}


	/**
	 * Set notification handler. Do noting.
	 */	
	public void setNotificationHandler(NotificationHandler nh)
	{
	}


	/** 
	 * Return the result of query for the statistics specified by the given xpath.
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance represents the statistics to be queried.
	 * @see I_Stats#getStats
	 */
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return null;
	}

	/** 
	 * Set the statistics specified by the given xpath. The value actually set is
	 * returned.
	 * 
	 * @param doc   	Document used for create new elements or nodes.
	 * @param xpath 	XPath instance represents the statistics to be queried.
	 * @param newValue	Element representing the value or sub-tree to be set.
	 * @see I_Stats#setStats
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		return null;
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in read schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return null;
	}

	/** 
	 * Return the schema element which represents the root of the sub-tree, specified
	 * by the given xpath, in write schema tree. 
	 * 
	 * @param doc   Document used for create new elements or nodes.
	 * @param xpath XPath instance representing the statistics which is the root
	 * 	of the sub-tree to be queried.
	 * @see I_Stats#getWriteSchema(Document, XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return null;
	}
	

	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return null;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
		/*
		 * Do nothing.
		 */
	}
}


/** Structure used to hold information on each node stored at the DT server. */
class NodeRecord 
{
    /** Maximum number of times the node will be queried by a new node.
      (Assuming this node is not the Leader.) */
    public static final int MAX_TIMES_PINGED = 6;  

    /** The timeout time for this record at the server. */
    public static final int TIMEOUT_TIME = 10000;  // in milliseconds

    /** The address pair of the node. */
    public DT_AddressPair dtap;
    /** The last time a message was received from the node. */
    public long lastContactTime;

    /** The first time a message was received from the node. */
    public long firstContactTime;

    /** The number of times this node has been pinged. */
    public int timesPinged;

    NodeRecord(DT_AddressPair ap) 
    {
        dtap = ap;
        firstContactTime = System.currentTimeMillis();
        lastContactTime = firstContactTime;
        timesPinged = 0;
    }

    public void touch() 
    {
        lastContactTime = System.currentTimeMillis();
    }
}

/** Data Structure for the node cache at the DT_Server */
class NodeCache 
{
    /** The maximum number of nodes stored for any Overlay. */
    final static int MAX_NODES = 300;

    /** The cache of nodes in the Overlay. */
    private Vector nodeCache;

    public NodeCache() 
    {
        nodeCache = new Vector();
    }

    public int size()
    {
        return nodeCache.size();
    }
    
    public NodeRecord elementAt(int i) 
    {
        return (NodeRecord) nodeCache.elementAt(i);
    }

    public NodeRecord getLeader()
    {
        if (nodeCache.size() == 0)
            return null;
        else
            return (NodeRecord) nodeCache.elementAt(0);
    }

    public int getLeaderIndex() 
    {
        return 0;
    }
    
    
    /** Removes all node records from a Vector that have timed out.
     * It then makes sure the node with the highest logical address is
     * moved to the front of the list. 
    */
    public void removeOldRecords() 
    {
        long t = System.currentTimeMillis();

        for (int i = 1; i < nodeCache.size(); i++) 
        {
            if ( t - ((NodeRecord) nodeCache.elementAt(i)).lastContactTime > NodeRecord.TIMEOUT_TIME) 
            {
                nodeCache.removeElementAt(i);
                i--;
            }
        }

 //      if leader times out...
      if (nodeCache.size() > 0 && t - ((NodeRecord) nodeCache.elementAt(0)).lastContactTime > NodeRecord.TIMEOUT_TIME)
      {
           System.out.println("DT Server: Leader replaced was selected " + getLeader().timesPinged + " times.");
             nodeCache.removeElementAt(0);
             calculateNewLeader();
      }

    }

    private void calculateNewLeader() 
    {
        if (nodeCache.size() == 0)
            return;
        
        int indexOfNewLeader = 0;
        DT_AddressPair newLeader = ((NodeRecord) nodeCache.elementAt(0)).dtap;

        for (int i = 1; i < nodeCache.size(); i++) 
        {
            DT_AddressPair possibleNewLeader = ((NodeRecord) nodeCache.elementAt(i)).dtap;
            if (((DT_LogicalAddress) possibleNewLeader.getLogicalAddress()).greaterThan((DT_LogicalAddress) newLeader.getLogicalAddress()))
            {
                newLeader = possibleNewLeader;
                indexOfNewLeader = i;
            }
        }

        // swap record at index of new leader with record at 0.
        NodeRecord tmp = (NodeRecord) nodeCache.elementAt(0);
        nodeCache.setElementAt(nodeCache.elementAt(indexOfNewLeader), 0);
        nodeCache.setElementAt(tmp, indexOfNewLeader);
        System.out.println("DT Server: New leader at coords: " + newLeader.getLogicalAddress().toString() + " at time " + System.currentTimeMillis());
    }

    public void update(DT_AddressPair addMe) 
    {
        NodeRecord record = getRecord(addMe.getPhysicalAddress());
        if (record != null) 
        {
            record.dtap = addMe;
            record.touch();
        }
        else 
        {
            NodeRecord leader = getLeader();
        
            if (leader == null ||
                addMe.getDTLogicalAddress().greaterThan(leader.dtap.getLogicalAddress())) 
            {
                NodeRecord newLeader = new NodeRecord(addMe);
//                 System.out.println("Leader replaced was selected " + ((leader == null)?0:leader.timesPinged) + " times.");
//                 System.out.println("New leader at coords: " + addMe.getLogicalAddress().toString() + " at time " + System.currentTimeMillis());
                if (nodeCache.size() == 0)
                {
                    nodeCache.addElement(newLeader);
                }
                else 
                {
                    nodeCache.addElement(leader);
                    nodeCache.setElementAt(newLeader, 0);

                    // if overfull, remove random element (not Leader).
                    if (nodeCache.size() >= MAX_NODES)
                        nodeCache.removeElementAt(1 + (Math.abs(DT_Server.rand.nextInt()) % (nodeCache.size()-1)));
                }
            }
            else if (nodeCache.size() < MAX_NODES)
            {
                nodeCache.addElement( new NodeRecord(addMe) );
            }
        }
    }
    
	/** get a record by the physical address */
    public NodeRecord getRecord(I_PhysicalAddress PA) 
    {
        for (int i = 0; i < nodeCache.size(); i++) 
        {
            if ( ((NodeRecord) nodeCache.elementAt(i)).dtap.getPhysicalAddress().equals(PA))
                return (NodeRecord) nodeCache.elementAt(i);
        }

        return null;
    }

	/** remove a node with the specified physical address */
    public void remove(I_PhysicalAddress PA) 
    {
        for (int i = 0; i < nodeCache.size(); i++) 
        {
            if ( ((NodeRecord) nodeCache.elementAt(i)).dtap.getPhysicalAddress().equals(PA))
            {
//                if (i == 0)
//                    System.out.println("Leader replaced was selected " + getLeader().timesPinged + " times.");
                    
                nodeCache.removeElementAt(i);
                if (i == 0)
                    calculateNewLeader();
                return;
            }
        }
    }
    

    /** Computes the node to be sent in a ServerReply message.
     * If the argument is the node that is the Leader, then the
     * Leader is returned.
     * Otherwise, the function returns the node closest to the argument
     * such that the returned node has coordinates greater than the argument.
    */
    public final static long TIME_TO_BE_OLD = 12000; // milliseconds
    public final static double MIN_PORTION_OLD = 0.5; 

    public DT_AddressPair getAParent(DT_AddressPair subject) 
    {
        if (nodeCache.size() == 0)  
            throw new IllegalStateException("Requested data from an empty NodeCache.");

        DT_AddressPair parent = null;
        NodeRecord parentRecord = null;
        int parentIndex = -1;

        DT_AddressPair parent_old = null;
        NodeRecord parentRecord_old = null;
        int parentIndex_old = -1;

        int numOfOldEntries = 0;

//         int indexOfMostPinged = -1;
//         long pingsOfMostPinged = -1;

        long time = System.currentTimeMillis();

        for (int i = 0; i < nodeCache.size(); i++) 
        {
            NodeRecord possibleParentRecord = (NodeRecord) nodeCache.elementAt(i);
            DT_AddressPair possibleParent = possibleParentRecord.dtap;

            if (time - possibleParentRecord.firstContactTime >= TIME_TO_BE_OLD) 
                numOfOldEntries++;
            
//             if (i != 0 && possibleParentRecord.timesPinged > pingsOfMostPinged) 
//             {
//                 indexOfMostPinged = i;
//                 pingsOfMostPinged = possibleParentRecord.timesPinged;
//             }
            
			//It is to ensure it works in the case that multiple nodes have the same logical addresses.
            //if (((DT_LogicalAddress) possibleParent.getLogicalAddress()).greaterThan((DT_LogicalAddress) subject.getLogicalAddress())
            if (!((DT_LogicalAddress) subject.getLogicalAddress()).greaterThan((DT_LogicalAddress) possibleParent.getLogicalAddress())
                && !(possibleParent.getPhysicalAddress().equals(subject.getPhysicalAddress())))
            {
                if (parent == null || 
                    (DT_Point.distance((DT_LogicalAddress) parent.getLogicalAddress(), (DT_LogicalAddress) subject.getLogicalAddress()) 
                     >= DT_Point.distance((DT_LogicalAddress) possibleParent.getLogicalAddress(), (DT_LogicalAddress) subject.getLogicalAddress()))) 
                {
                    parent = possibleParent;
                    parentRecord = possibleParentRecord;
                    parentIndex = i;
                }  


                if (time - possibleParentRecord.firstContactTime >= TIME_TO_BE_OLD &&
                    (parent_old == null || 
                     (DT_Point.distance((DT_LogicalAddress) parent_old.getLogicalAddress(), (DT_LogicalAddress) subject.getLogicalAddress()) 
                      >= DT_Point.distance((DT_LogicalAddress) possibleParent.getLogicalAddress(), (DT_LogicalAddress) subject.getLogicalAddress()))))
                {
                    parent_old = possibleParent;
                    parentRecord_old = possibleParentRecord;
                    parentIndex_old = i;
                }
            }
        }
        
//        System.err.println("#### total = " + nodeCache.size() + "     # old = " + numOfOldEntries );

        // remove node if necessary.
        // (Assume the most pinged is "old")
//         if (nodeCache.size() == MAX_NODES && numOfOldEntries > MAX_NODES*MIN_PORTION_OLD) 
//         {
//             double probabilityToRemoveNode = (numOfOldEntries - MAX_NODES*MIN_PORTION_OLD)/(MAX_NODES - MAX_NODES*MIN_PORTION_OLD);
//             if (GroupObject.rand.nextDouble() < probabilityToRemoveNode) 
//             {
//                 System.err.println("Removed node.  # of pings = " + pingsOfMostPinged + "    P=" + probabilityToRemoveNode);
//                 nodeCache.removeElementAt(indexOfMostPinged);
//             }
//         }
        

        if (parent != null && numOfOldEntries < nodeCache.size()*MIN_PORTION_OLD && parent != parent_old)
        {
//            System.err.println("\tSelected young node.");
             parentRecord.timesPinged++;
             if (parentIndex != 0 && parentRecord.timesPinged >= NodeRecord.MAX_TIMES_PINGED) 
             {
//               System.err.println("\tRemoved node.");
                 nodeCache.removeElementAt(parentIndex);
             }
             return parent;
        }
        else if (parent_old != null)
        {
//            System.err.println("\tSelected old node.");
            parentRecord_old.timesPinged++;
            if (parentIndex_old != 0 && parentRecord_old.timesPinged >= NodeRecord.MAX_TIMES_PINGED) 
            {
//                System.err.println("\tRemoved node.");
                nodeCache.removeElementAt(parentIndex_old);
            }
            return parent_old;
        }
        else 
        {
//           System.err.println("\tSelected Leader.");
            NodeRecord leader = getLeader();
            leader.timesPinged++;
            return leader.dtap;
        }
    }
}
