/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

/**
 * This is the interface for application messages.  It is implemented
 * by the class OL_Message.
 * 
 * @author HyperCast Team
 * @author Dongwen Wang
 * @version 2.0, Feb. 23, 2001
 * @author Jianping Wang
 * @version 3.0, Jul. 11, 2004
 */

public interface I_OverlayMessage
{
	/*
	 * The following conatants are equivalent to the FSM ID
	 * constants defined in MessageStore class.
	 */
	/**
	 * Service identifiers for enhanced service messages.
	 */
	public static final short NOSERVICE			= MessageStore.FSM_NO_SERVICE; //0
	public static final short H2HACK			= MessageStore.FSM_H2H_ACK; //1
	public static final short E2EACK			= MessageStore.FSM_E2E_ACK; //2
	public static final short DELDUPS			= MessageStore.FSM_DUPLICATE_ELIMINATION; //3
	public static final short SYNC				= MessageStore.FSM_SYNCHRONIZATION; //4
	public static final short INCAST			= MessageStore.FSM_INCAST; //5
	public static final short INORDER			= MessageStore.FSM_BEST_EFFORT_ORDERING; //6
	public static final short NAMING			= MessageStore.FSM_NAMING; //7
	
    /**
     * Returns the maximum number of hops that this message may take
     * in the network.  If called on a message that has been received
     * from some other node this value is the original hop limit minus
     * the number of hops that the message has already taken.
     */
    public short getHopLimit();

    /**
     * Sets the hop limit for only this message.  The hop limit is the
     * maximum number of hops this message will cross before being
     * discarded.  The default hop limit for all messages can be set
     * by calling <code>I_OverlaySocket.setHopLimit</code>.
     *
     * 0 <= hop limit <= (2^16) - 1
     * The hopLimit is interpretted as an <em>unsigned</em> two byte quantity.
     */
    public void setHopLimit (short hopLimit);

    /**
     * Returns the payload byte array carried by this message.
     * Only <code>getPayloadLength()</code> bytes of the payload are valid. 
     */
    public byte[] getPayload();

    /**
     * Returns the number of valid bytes in the payload.
     */
    public int getPayloadLength();

    /**
     * Returns the logical address of the sender of this message.
     * @deprecated use getSourceAddress();
     */
    public I_LogicalAddress getSrcAddress();
    public I_LogicalAddress getSourceAddress();

    /** 
     * Returns the traffic class of the message
     */
    public byte getTrafficClass();

    /** 
     * Sets the traffic class of the message. 
     */
    public void setTrafficClass (byte trafficclass);

    /**
     * Returns the unique identifier of this message.
     * 
     * @deprecated use getMessageIdentifier
     */
    public byte[] getMessageID();

    /**
     * Returns the service number carried by this message.
     * 
     * @deprecated use getServiceIdentifier
     */
    public short getServiceNumber();

    /** 
     * Gets the stream ID of the message. This method is useful only when 
     * messageStore is used for the transmission of application messages.
     * If there is no stream id associated with this message, null is returned.
	 * 
     * @deprecated use getStreamIdentifier
     */
    public byte[] getStreamID ();
    

	/**
	 * Returns the service number carried by this message. It returns -1
	 * if the message is not a enhanced service message.
	 * 
	 */
	public short getServiceIdentifier();

	/**
	 * Sets the service number in this message if the message is an enhanced service message.
	 */
	public void setServiceIdentifier(short serviceID);

	/**
	 * Returns the stream identifier carried by this message. It returns -1
	 * if the message is not a enhanced service message.
	 */
	public int getStreamIdentifier();

	/**
	 * Sets the stream identifier in this message if the message is an enhanced service message.
	 */
	public void setStreamIdentifier(int streamID);
	
	/**
	 * Returns the sequence number carried by this message. It returns -1
	 * if the message is not a stream-oriented message.
	 */
	public long getSequenceNumber();

	/**
	 * Sets the sequence number in this message if the message is an stream-oriented message.
	 */
	public void setSequenceNumber(long sequenceNum);
	
	/**
	 * @return the message identifier carried by this message. It returns -1
	 * if the message is not a enhanced service message.
	 */
	public byte [] getMessageIdentifier();
	
	/**
	 * Sets the message identifier in this message if the message is an enhanced service message.
	 */
	public void setMsgIdentifier(int msgID);
    
}
