/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import hypercast.events.NOTIFICATION_EVENT;
import hypercast.I_ReceiveCallback;


/**
 * This class is a finite state machine which handles best effort ordering for stream.
 * 
 * @author HyperCast Team
 * @author Guimin Zhang
 * @version 2.0, May. 23, 2001 */


public class MessageStoreFSM_Ordering extends I_MessageStoreFSM
{
		/** OL_Socket object*/
	private OL_Socket socket;
		/** MessageStore*/
	private MessageStore SMStore;
		
		/** The state information of the message */
	private byte MyState = Init;
		/** current sequence Number */
	private long currentSequenceNumber = 0;
		/**ADF_Payload for Ordering*/
	public static final byte PAYLOAD		= -1;
		/** The state of the stream,Init */
	public static final byte Init = 0;
		/** The state of the stream,Ordering */
	public static final byte Ordering = 1;

    private static FiniteStateMachineMetaOperations metaOps;
    protected static FiniteStateMachineMetaOperations registerMetaOperations ()
    {
        if (metaOps == null)
        {
            metaOps = new OrderingMetaOperations();
        }
        return metaOps;
    }

	/**
	 * Constructs a MessageStoreFSM_Ordering object with OL_Socket and MessageStore. 
	 */
	public MessageStoreFSM_Ordering(OL_Socket socket,MessageStore SMstore,HyperCastConfig config)
	{
		/*
		 * Initialization
		 */
		this.socket = socket;
		this.SMStore = SMstore;
	}
	/** check the change of the neighborhood
	 */	
	public void changingNeighbor()
	{
		 //no body needed
	}  
	
	/**
	 * @return the finite state machine id
	 */
	public short getFSMID()
	{
		return MessageStore.FSM_BEST_EFFORT_ORDERING;
	}
	/**process new ADF_Data message
	 * @param	msg			OL_Message
	 */
	public void newADMsg(OL_Message msg)
	{
        /*
         * pass to the streammanager if it is registered as a callback
         */
        I_ReceiveCallback callback = (I_ReceiveCallback) MessageStore.fsmCallbacks.get (new Short (getFSMID()));
        

		if ((msg.getDeliveryMode() == OL_Message.DELIVERY_MODE_UNICAST) && (!(socket.getLogicalAddress().equals(msg.getDestinationAddress()))))
		{
			return;
		}

        if (callback != null)
        {

// TODO: The following should be logged and/or displayed in debug mode
			//System.out.println("MessageStoreFSM_Ordering: newADMsg: find callback.");

			//if ((msg.getDeliveryMode() == OL_Message.DELIVERY_MODE_UNICAST) && (!(socket.getLogicalAddress().equals(msg.getDestinationAddress()))))
			//{
			//	return;
			//}

			socket.appmsgArrived(msg,callback);
        }
        /*
         * forward the message to the application
         */
        else
        {
			socket.appmsgArrived(msg,socket.callback);
        }
		
		/*
		 * change the state to Ordering
		 */
		MyState = Ordering;
		/*
		 * extract the ADF header information
		 */
		FSM_Extension e = (FSM_Extension) msg.getFirstExtensionByType(Extension.FSM);
		/*
		 * update the current sequence number
		 */
		currentSequenceNumber = e.getSequenceNumber();

	}

	/**Process the new ADF_Control message 
	 * @param	msg			OL_Message
	 */
	public void newACMsg(OL_Message msg)
	{
		//no body needed	
	}	
	
    /**Update the known ADF_Data message entry
     * @param	msg			OL_Message
     */			
    public void updateADMsg (OL_Message msg)
    {
        /*
         * extract the ADF header information
         */
        FSM_Extension e = (FSM_Extension) msg.getFirstExtensionByType(Extension.FSM);
        /*
         * get the sequence number of the message
         */	
        long seq = e.getSequenceNumber();
        
        /*
         * the sequence number did not exceed the maxium value
         */
        if(Math.abs(seq - currentSequenceNumber) < FSM_Extension.MAX_SEQUENCE_NUMBER_VALUE/2) 
        {
			/*
			 * discard the out of order message
			 */
			if(seq <= currentSequenceNumber)
			{
				return;
			}
			/*
			 * for in sequence message
			 */
			else if(seq > currentSequenceNumber)
			{
				if ((msg.getDeliveryMode() == OL_Message.DELIVERY_MODE_UNICAST) && (!(socket.getLogicalAddress().equals(msg.getDestinationAddress()))))
				{
					return;
				}

				/*
				 * pass to the sreammanager if it is not null
				 */
				I_ReceiveCallback callback = (I_ReceiveCallback) MessageStore.fsmCallbacks.get (new Short (getFSMID()));
				if (callback != null)
				{
					socket.appmsgArrived(msg,callback);
				}
				/*
				 * forward the message to the application
				 */
				else
				{
					socket.appmsgArrived(msg,socket.callback);
				}    
				
				/*
				 * update current sequence number
				 */            
				 currentSequenceNumber = seq;
			}
        }
        else
        {
			/*
			 * discard the out of order message
			 */
			if(seq >= currentSequenceNumber)
			{
				return;
			}
			/*
			 * for in sequence message
			 */
			else if(seq < currentSequenceNumber)
			{
				/*
				 * pass to the sreammanager if it is not null
				 */
				I_ReceiveCallback callback = (I_ReceiveCallback) MessageStore.fsmCallbacks.get (new Short (getFSMID()));
				if (callback != null)
				{
					socket.appmsgArrived(msg,callback);
				}
				/*
				 * forward the message to the application
				 */
				else
				{
					socket.appmsgArrived(msg,socket.callback);
				}    
				
				/*
				 * update current sequence number
				 */            
				 currentSequenceNumber = seq;
			}	
        }



    }
	
	/**Update the new ADF_Control message 
	 * @param	msg			OL_Message
	 */
	public void updateACMsg(OL_Message msg)
	{
		//nothing
	}


		
	/**Set the time expired events
	 * @param	timerIndex		
	 */
	public synchronized void timerExpired(int timerIndex)
	{
		//no body needed
	}
}

class OrderingMetaOperations implements FiniteStateMachineMetaOperations
{
    public boolean messageStoreWillForwardMessage (OL_Message message) { return false; }
    public boolean isValidMessage (OL_Message message) { return true; }
	//public boolean processIntermediateUnicastMessage(OL_Message message){return true;}
	public boolean processIntermediateUnicastMessage(OL_Message message){return false;}
    public NOTIFICATION_EVENT notifyOnSend (OL_Message message) { return null; }
}
