/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.MonitorAndControl;

import hypercast.I_Stats;
import hypercast.HyperCastWarningRuntimeException;

import java.io.FileNotFoundException;
/**
 * @see Portal
 * 
 * @author HyperCast Team
 * @author Guangyu Dong, January 2005
 * @author Greg Mattes, January 2005
 */
public interface I_Portal
    extends I_Stats
{
    /** 
     * Opens the <code>Portal</code>. This causes the
     * <code>PortalAdvertiser</code> thread to unblock so that
     * <code>MCAdvertiseMessages</code> begin to flow from the
     * <code>Portal</code> to the <code>Monitor</code>.
     * 
     * A <code>Portal</code> cannot be used until it is opened.
     */
    public void openPortal();

    /**
     * Closes the portal. This causes the
     * <code>PortalAdvertiser</code> thread to block so that
     * <code>MCAdvertiseMessages</code> cease to flow from the
     * <code>Portal</code> to the <code>Monitor</code>.
     */
    public void closePortal();

    /** 
     * Causes <code>MCMessages</code> to be recorded in a transcript
     * file. Depending on the options present in the configuration
     * file either messages that are sent, received, or both will be
     * recorded.
     *
     * @param filename The name of the transcript file.
     *
     * @throws HyperCastWarningRuntimeException If the header cannot be written to the transcript file.
     * @throws HyperCastWarningRuntimeException If the configuration
     * specifies that neither sent nor received messsages should be
     * recorded.
     * @throws FileNotFoundException If the specified file cannot be accessed
     * @throws IllegalStateException If a transcript is already started
     */
    public void startTranscript (final String filename)
        throws FileNotFoundException, IllegalStateException, HyperCastWarningRuntimeException;
    
    /**
     * Stops transcripting.
     *
     * @throws IllegalStateException If no transcript is active.
     */
    public void stopTranscript()
        throws IllegalStateException, HyperCastWarningRuntimeException;

    /**
     * Causes a Portal to cease sending MonitorAdvertise messages to a Monitor
     */
    public void setSilentOn();

    /**
     * Causes a Portal to begin sending MonitorAdvertise messages to a Monitor
     */
    public void setSilentOff();
}
