/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.MonitorAndControl;

/**
 * The application using the Monitor creates an instance of this
 * interface and passes it as a parameter when sending out a
 * <code>MonMessage</code> by using the <code>sendTo()</code> method
 * defined in the <code>I_Monitor</code> interface, and specify a
 * timeout value for waiting for the reply message.
 *
 * If during the specified timeout time there is no reply message coming
 * back, the method <code>timeoutMCMessage()</code> will be called by the Monitor.
 * 
 * The <code>Timeout_Callback</code> is managed by the
 * <code>Monitor</code> <code>CallbackManager</code> which stores it
 * in a <code>CallbackEntry</code>.
 *  
 * @author HyperCast Team
 * @author Guangyu Dong, January 2005
 * @author Greg Mattes, January 2005
 */
public interface I_TimeoutCallback
{
    /**
     * Handling timeout for a <code>MonMessage</code>. The handler can
     * choose to resend the message; resending is implemented by a
     * higher layer - i.e. no in the <code>Monitor</code>.
     * 
     * @param message Message for which there was no reply in the
     * specified timeout time.
     */
    public void timeoutMessage (final MonMessage message);
}
