/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.NONE;

import hypercast.I_Address;
import hypercast.I_LogicalAddress;
import hypercast.I_PhysicalAddress;

/**
 * 
 * @author HyperCast Team
 * @author Guangyu Dong
 * @version 2005 (version 3.0)
 */
public class NONE_LogicalAddress
    implements I_LogicalAddress
{
    private I_PhysicalAddress physicalAddress;
    private int logicalAddressSize;
        
    // TODO: so... one constructor guarentees that the logical address
    // size is larger than the physical address size, but the other
    // constructor *ignores* logical address size altogether?
    private String logicalAddressString;
	
    //to create a physical address only for display
    //without a node
    public NONE_LogicalAddress (final String logicalAddressString)
    {
        this.physicalAddress = null;
        this.logicalAddressString = logicalAddressString;
        // TODO: this.logicalAddressSize ???
    }

    public NONE_LogicalAddress (final I_PhysicalAddress physicalAddress)
    {
        this.physicalAddress = physicalAddress;

        final int physicalAddressSize = physicalAddress.getSize();
        this.logicalAddressSize = 1;
        while (logicalAddressSize < physicalAddressSize)
        {
            this.logicalAddressSize *= 2;
        }
        // TODO: this.logicalAddressString ???
    }

    public I_PhysicalAddress getPhysicalAddress()
    {
        return physicalAddress;
    }

    /**
     * @see hypercast.I_Address#greaterThan(hypercast.I_Address)
     */
    public boolean greaterThan (I_Address a)
    {
        return false;
    }

    /**
     * @see hypercast.I_Address#toByteArray()
     */
    public byte [] toByteArray()
    {
        final byte [] result = new byte [logicalAddressSize];
        System.arraycopy (physicalAddress.toByteArray(), 0, result, 0, physicalAddress.getSize());
        return result;
    }

    /**
     * @see hypercast.I_Address#getSize()
     */
    public int getSize()
    {
        return logicalAddressSize;
    }

    public String toString()
    {
        return (physicalAddress != null) ? physicalAddress.toString() : logicalAddressString;
    }

    public boolean equals (final Object object)
    {
        boolean result = false;
        if (object instanceof NONE_LogicalAddress)
        {
        	if(physicalAddress==null)
        		result = toString().equals(object.toString());
        	else{
        		final NONE_LogicalAddress other = (NONE_LogicalAddress) object;
        		result = this.physicalAddress.equals (other.physicalAddress);
        	}
        }
        return result;
    }
}
