/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/
package hypercast;

import hypercast.util.XmlUtil;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

/**
 * This class implements methods to create a new node based on the
 * specifications in the configuration object.
 *
 * @author HyperCast Team
 * @author Jianping Wang
 * @version 3.0, Mar. 15, 2004
 * @author Greg Mattes
 * @version 3.0, Spring 2005 - reflective instanciation of Pastry node that allows decoupling of HC and Pastry jars and compilation processes
 * @author Wittawat Tantisiriroj
 * @version Jul. 03, 2005 
*/
public final class NodeFactory {
	
	/**
	 * Constructor.
	 */
	public NodeFactory() {
	}
	
	/**
	 * This method creates I_Node object based on the node types specified in
	 * the configuration object.
	 *
	 * @param config socket configuration object.
	 * @param ss node adapter that is passed to the node constructor.
	 *
	 * @return an I_Node object.
	 */
	public I_Node create_newNode(HyperCastConfig config, I_UnicastAdapter ss) {
		return create_newNode(config, ss, "/Public/Node");
	}
	/**
	 * This method creates I_Node object based on the node types specified in
	 * the configuration object.
	 *
	 * @param config socket configuration object.
	 * @param ss node adapter that is passed to the node constructor.
	 *
	 * @return an I_Node object.
	 */
	public I_Node create_newNode(HyperCastConfig config, I_UnicastAdapter ss, String prefix) {
		I_NodeLoader nodeLoader = null;
		I_Node node = null;

		String nodeSelection =
			config.getTextAttribute (XmlUtil.createXPath(prefix));

		final String nodeClassName = "hypercast.nodeloaders."+nodeSelection;
		if (nodeClassName != null) {
			final String errorString = "\""
					+ nodeSelection
					+ "\" is not supported - you must install the separate jar file for HyperCast";
			try {
				final Class nodeClass = Class.forName(nodeClassName);
				final Constructor constructor = nodeClass.getConstructor(new Class[] {});
				final Object[] constructorParameterValues = new Object[] {};
				nodeLoader = (I_NodeLoader) constructor
						.newInstance(constructorParameterValues);
				node = nodeLoader.create_newNode(config, ss, prefix+ "/" + nodeSelection + "/");
			} catch (final ClassNotFoundException cnfe) {
				throw new HyperCastConfigException(errorString, cnfe);
			} catch (final ClassCastException cce) {
				throw new HyperCastConfigException("NodeFactory: Error: Node property is wrong!", cce);
			} catch (final NoSuchMethodException nsme) {
				throw new HyperCastConfigException("NodeFactory: Error: Node property is wrong!", nsme);
			} catch (final InstantiationException ie) {
				throw new HyperCastConfigException("NodeFactory: Error: Node property is wrong!", ie);
			} catch (final IllegalAccessException iae) {
				throw new HyperCastConfigException("NodeFactory: Error: Node property is wrong!", iae);
			} catch (final InvocationTargetException ite) {
				throw new HyperCastConfigException("NodeFactory: Error: Node property is wrong!", ite);
			}
		} else {
			throw new HyperCastFatalRuntimeException(
					"NodeFactory: Error: Node property is wrong!");
		}
		
		return node;
	}
}
