/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.Pastry;

import hypercast.*;
import java.io.*;
import java.util.*;
import java.net.*;

/** Structure used to hold information of each buddy.
 * @author Weisheng Si
 * @version 2.0, Nov 11, 2001 
 *
 * @author HyperCast Team
 * @author Greg Mattes
 * @version May 2004
*/
class BuddyRecord 
{
    private final Pastry_AddressPair buddy;

    public BuddyRecord (final Pastry_AddressPair buddy) {
        this.buddy = buddy;
    }

    /**
     * constructor of BuddyRecord from a string.  the string has to be
     * in syntax [<la>@]pa, where the logical address part is
     * optional.
     */
    public BuddyRecord (final I_UnicastAdapter adapter, final String addressPairString)
    {
        final Pastry_LogicalAddress la;
        final String paString;

        final int i = addressPairString.indexOf ('>');
        if (i >= 0)
        {
            final String nodeIdStr= addressPairString.substring (addressPairString.indexOf ('<') + 1, i);
            la = new Pastry_LogicalAddress (nodeIdStr);
            paString = addressPairString.substring (i + 2, addressPairString.length()); // skip '>' and '@'
        }
        else
        {
            la = null;
            paString = addressPairString.substring (0, addressPairString.length());
        }

        try
        {
            I_PhysicalAddress pa = adapter.createPhysicalAddress (paString);
            this.buddy = new Pastry_AddressPair (pa, la);
        }
        catch (IllegalArgumentException e)
        {
            throw new HyperCastConfigException ("Buddy address is invalid: " + addressPairString, e);
        }
    }

    public Pastry_AddressPair getAddressPair() {
        return buddy;
    }

    /** Creates a string following the "(la)@pa" syntax. */
    public String toString() {
        final Pastry_LogicalAddress la = (Pastry_LogicalAddress) buddy.getLogicalAddress();
        return (la != null ? "(" + la + ")@" : "") + buddy.getPhysicalAddress();
    }   
   
    /** 
     * Compares two Buddy records.
     * If they have the same physical address, then we believe they are the same buddy.
     * It is important that the argument is of type <code>Object</code> so that 
     * it can be used in a <code>Hashtable</code>.
     * 
     * @param    obj an object (should be of type <code>BuddyRecord</code>.)
     * @exception IllegalArgumentException if <code>obj</code> is not of type 
     * <code>BuddyRecord</code>.
     */
    public boolean equals (final Object obj)
    {
        if (obj == null)
            return false;
    
        if (! (obj instanceof BuddyRecord))
            throw new IllegalArgumentException ("Error in equals(): not of type BuddyRecord!");

        final BuddyRecord other = (BuddyRecord) obj;
        return buddy.getPhysicalAddress().equals (other.getAddressPair().getPhysicalAddress());
    }
}
