/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast;

import java.io.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

import org.apache.xpath.*;


/**
 * This class define a common class which implements I_Stats. Any I_Stats class
 * for a scalar statistic must extend this class.
 * 
 * @author HyperCast Team
 * @author Jianping
 * @version	1.0, Jan. 25, 2005
 */

public abstract class SimpleStats implements I_Stats
{
	/**
	 * The statistics name for the object (as an I_Stats instance) which implements this class.
	 * It must be set before the the object can corectly process statistics query. 
	 * 
	 * @see I_Stats interface.
	 */
	protected String statisticsName = null;

    /** 
     * Constructor.
     */
    public SimpleStats ()
    {
    }

	/** @see I_Stats#getStatsName() */
	public String getStatsName ()
	{
		return statisticsName;
	}

	/** @see I_Stats#setStatsName(String) */
	public void setStatsName (String name)
	{
		statisticsName = name;
	}


	/** @see I_Stats#getStats(Document, XPath) */
	public Element[] getStats (final Document doc, final XPath xpath) throws HyperCastStatsException
	{
		String resultStr = getStats();
		if ((resultStr == null) || (statisticsName == null))
		{
			return new Element[0];
		}
		
		Element[] result = new Element[1];
		result[0] = doc.createElement(statisticsName);
		result[0].appendChild(doc.createTextNode(resultStr));

		return result;
	}

	/** 
	 * @see I_Stats#setStats(Document, XPath, Element)
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException
	{
		/*
		 * The format of newValue here can onle be 
		 *  <name>      //name must be the same as the variable statisticsName in this class.
		 *    <value/>  //TEXT_NOde
		 *  </name>
		 */
		if ((statisticsName == null) || !statisticsName.equals(newValue.getNodeName()))
		{
			//throw new HyperCastStatsException(xpath.getPatternString(), HyperCastStatsException.IncorrectSetValue);
			return new Element[0];
		}

		NodeList nl = newValue.getChildNodes();
		
		if ((nl.getLength() != 1) || (nl.item(0).getNodeType() != Node.TEXT_NODE))
		{
			//throw new HyperCastStatsException(xpath.getPatternString(), HyperCastStatsException.IncorrectSetValue);
			return new Element[0];
		}
		
		String newValueStr = nl.item(0).getNodeValue();
		String returnValue = setStats(newValueStr);

		if (returnValue == null)
		{
			return new Element[0];
		}
		else
		{
			Element[] result = new Element[1];
			result[0] = doc.createElement(statisticsName);
		
			result[0].appendChild(doc.createTextNode(returnValue));

			return result;
		}
	}

	/** 
	 * @see I_Stats#getReadSchema(Document, XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return new Element[0];
	}

	/** 
	 * @see I_Stats#getWriteSchema(Document, XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException
	{
		return new Element[0];
	}
	
	/**
	 * Gets the value of a scalar statistic managed by the implementation of this class.
	 * For a readable statistic, this method is overrided in the extended class.
	 */
	protected String getStats() throws HyperCastStatsException
	{
		return null;
	}
	
	/**
	 * Sets the value of a scalar statistic managed by the implementation of this class.
	 * For a writable statistic, this method is overrided in the extended class.
	 */
	protected String setStats(String newvalue) throws HyperCastStatsException
	{
		return null;
	}
}






