/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package hypercast.util;
import hypercast.util.XmlUtil;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.w3c.dom.Node;

import java.io.InputStream;
import java.util.Vector;
import java.io.IOException;


/**
 * This class provides functions to create a schema tree for
 * the given configuration XML document based on the hypercast
 * configuration schema.
 *
 * @author HyperCast Team
 * @version 2005 (version 3.0)
 */
public class CurrentConfigurationSchemaGenerator
{
	/** The Document instance created from the opened schema file. */
	private Document openedSchemaFile = null;


	/**
	 * Constructor.
	 */
	public CurrentConfigurationSchemaGenerator()
	throws IOException
	{
		String schemaFileName = "hypercast.xsd";

		InputStream fis = getClass().getResourceAsStream( schemaFileName );
				
		if (fis == null)
		{
			System.out.println("Connot find default schema file "+schemaFileName+".");
			return;
		}
				
		/*
		 * Open schema file.
		 */
		try
		{ 
			openedSchemaFile = XmlUtil.createDocument(fis, new String("hypercast.xsd"));
		}
		catch (final IOException ioe)
		{
			throw new IOException("Cannot open schema file "+schemaFileName+": "+ioe);
		}
				
		if (openedSchemaFile == null)
		{
			throw new IOException("Poor formated schema file: hypercast.xsd");
		}
		
	}

	/**
	 * Create schemaa for given configuration document.
	 */
	public Element cerateSchemaTree(Document doc, Document xmlConfigurationDoc)
	{
		try
		{
			/*
			 * Find the root Element of schema tree.
			 */
			Element propertyElement = createPrivatePropertyDocFromSchemaFile.findPropertyElement(openedSchemaFile, new String("Public"));
			if (propertyElement == null)
			{
				return null;						
			}

			/*
			 * Get the root Element in the given configuration document.
			 */
			Element xmlRoot = (Element) xmlConfigurationDoc.getChildNodes().item(0);

			if (!xmlRoot.getNodeName().equals(propertyElement.getAttribute("name")))
			{
				return null;
			}

			Element newSchemaRoot = doc.createElement("xs:element");
			newSchemaRoot.setAttribute("name", propertyElement.getAttribute("name"));
			
			createSchemaRecursive(propertyElement, xmlRoot, newSchemaRoot);

			return newSchemaRoot;
		}
		catch (Exception e)
		{
			System.err.println("Cannot create schema document for given XML configuration document.");
			e.printStackTrace();
			return null;
		}
	}

	/**
	 * Create schema subtree from given schema Element and configuration XML Element.
	 */
	public void createSchemaRecursive(Element schemaElement, Element xmlElement, Element newSchemaElement)
	{
		/*
		 * ComplexType Element
		 */
		if (createPrivatePropertyDocFromSchemaFile.isComplexElement(schemaElement))
		{
			//System.out.println("CurrentConfigurationSchemaGenerator:createSchemaRecursive: process complexType Element "+xmlElement.getNodeName()+".");

			Document doc = newSchemaElement.getOwnerDocument();
			Element complexType = doc.createElement("xs:complexType");
			newSchemaElement.appendChild(complexType);
			Element sequenceEle = doc.createElement("xs:sequence");
			complexType.appendChild(sequenceEle);

			
			NodeList firstnl = xmlElement.getChildNodes();

			Vector processedNames = new Vector();
			
			for (int i=0; i<firstnl.getLength(); i++)
			{
				Node node = firstnl.item(i);
				String name = firstnl.item(i).getNodeName();
				
				if (CurrentConfigurationSchemaGenerator.containName(processedNames, name))
					continue;
				
				Element matchEle = CurrentConfigurationSchemaGenerator.getChildOfComplexElement(schemaElement, name);
				if (matchEle != null)
				{
					Element newEle = (Element) doc.importNode(matchEle, false);
					sequenceEle.appendChild(newEle);
					
					createSchemaRecursive(matchEle, (Element) node, newEle);
				}
				else
				{
					throw new IllegalArgumentException("Element "+name+" in the configuration document doesn't find matched schema Element!");
				}
				
				processedNames.add(name);
			}
		}
		/*
		 * SimpleType Element.
		 */
		else
		{
			//System.out.println("CurrentConfigurationSchemaGenerator:createSchemaRecursive: process simpleType Element "+xmlElement.getNodeName()+".");

			Document doc = newSchemaElement.getOwnerDocument();

			Element restrictEle = createPrivatePropertyDocFromSchemaFile.getRestrictionElement(schemaElement);
			if (restrictEle != null)
			{
				Element simpleType = doc.createElement("xs:simpleType");
				newSchemaElement.appendChild(simpleType);

				if ((restrictEle.getAttribute("base").equals("IPv4AndOnePort"))
					|| (restrictEle.getAttribute("base").equals("IPv4AndTwoPorts")))
				{
					restrictEle.setAttribute("base", "xs:string");
				}

				Element newEle = (Element) doc.importNode(restrictEle, true);
				simpleType.appendChild(newEle);
			}
			else
			{
				newSchemaElement.setAttribute("type", "xs:String");	
			}
		}
	}

	/**
	 * Check if there is one child of the schema Element has the given name.
	 */
	private static boolean isOneChildOfComplexElement(Element schemaElement, String name)
	{
		Vector children = createPrivatePropertyDocFromSchemaFile.getChildrenOfComplexElement(schemaElement);
	
		for (int i=0; i<children.size(); i++)
		{	
			Element ele = (Element) children.elementAt(i);
			if (name.equals(ele.getAttribute("name")))
				return true;
		}
		return false;
	}

	/**
	 * Find the child of a ComplexType Node which has the given name.
	 */
	private static Element getChildOfComplexElement(Element schemaElement, String name)
	{
		Vector children = createPrivatePropertyDocFromSchemaFile.getChildrenOfComplexElement(schemaElement);
	
		for (int i=0; i<children.size(); i++)
		{	
			Element ele = (Element) children.elementAt(i);
			if (name.equals(ele.getAttribute("name")))
				return ele;
		}
		return null;
	}

	/**
	 * Check if the given name is contained in the Vector;
	 */
	private static boolean containName(Vector nameList, String name)
	{
		for (int i=0; i<nameList.size(); i++)
		{
			String nn = (String) nameList.elementAt(i);
			if (name.equals(nn))
				return true;
		}
		return false;
	}

}
